@@ -1,6 +1,6 @@
 /* io.c
  *
- * $Id: io.c,v 1.13 1999/06/09 14:56:08 bazsi Exp $ */
+ * $Id: io.c,v 1.15 1999/07/10 13:23:09 bazsi Exp $ */
 
 /* lsh, an implementation of the ssh protocol
  *
@@ -24,6 +24,7 @@
 #include "io.h"
 #include "werror.h"
 #include "xalloc.h"
+#include "format.h"
 
 #include <assert.h>
 #include <string.h>
@@ -46,6 +47,8 @@
 #include <fcntl.h>
 #include <sys/types.h>
 #include <sys/socket.h>
+#include <sys/stat.h>
+#include <sys/un.h>
 #include <arpa/inet.h>
 #include <signal.h>
 
@@ -473,20 +476,21 @@ static void write_callback(struct nonblocking_fd *fd)
 static void listen_callback(struct nonblocking_fd *fd)
 {
 	CAST(listen_fd, self, fd);
-	struct sockaddr_in peer;
-	size_t addr_len = sizeof(peer);
+	char sockaddr_buf[256];
+	struct sockaddr *peer = (struct sockaddr *) &sockaddr_buf;
+	size_t addr_len = sizeof(sockaddr_buf);
 	int res;
 	int conn;
 
 	conn = accept(fd->fd,
-		      (struct sockaddr *) &peer, &addr_len);
+		      peer, &addr_len);
 	if (conn < 0)
 	{
 		werror("io.c: accept() failed, %z", strerror(errno));
 		return;
 	}
 	res = FD_LISTEN_CALLBACK(self->callback, conn,
-				 addr_len, (struct sockaddr *) &peer);
+				 sockaddr2address_info(addr_len, peer));
 	if (ST_CLOSEDP(res))
 	{
 		closekill_fd(fd, (ST_FAILUREP(res)
@@ -498,14 +502,22 @@ static void listen_callback(struct nonblocking_fd *fd)
 static void connect_callback(struct nonblocking_fd *fd)
 {
 	CAST(connect_fd, self, fd);
-	int res;
+	int sock_err;
+	size_t len = sizeof(sock_err);
   
-	res = FD_CALLBACK(self->callback, fd->fd);
+  	if ((getsockopt(fd->fd, SOL_SOCKET, SO_ERROR, (char *) &sock_err, &len) < 0) ||
+  	    sock_err) {
+  	    	verbose("connect_callback(): connect() failed\n");
+  		FD_CALLBACK(self->callback, -1);
+  	}
+  	else {
+		int res = FD_CALLBACK(self->callback, fd->fd);
 
-	if (!ST_CLOSEDP(res))
-	{
-		/* To avoid actually closing the fd */
-		fd->fd = -1;
+		if (!ST_CLOSEDP(res))
+		{
+			/* To avoid actually closing the fd */
+			fd->fd = -1;
+		}
 	}
 	kill_fd(fd);
 }
@@ -712,6 +724,37 @@ get_inaddr(struct sockaddr_in	* addr,
 	return 1;
 }
 
+/* Converts a string port number or service name to a port number.
+ * Returns the port number in _host_ byte order, or -1 of the port
+  * or service was invalid. */
+  
+int get_portno(const char *service, const char *protocol)
+{
+  if (service == NULL)
+    return 0;
+  else
+    {
+      char *end;
+      long portno;
+
+      portno = strtol(service, &end, 10);
+      if (portno > 0
+          &&  portno <= 65535
+          &&  end != service
+          &&  *end == '\0')
+          return portno;
+      else
+        {
+          struct servent * serv;
+
+          serv = getservbyname(service, protocol);
+          if (serv == NULL)
+            return -1;
+          return ntohs(serv->s_port);
+        }
+    }
+}
+
 /* FIXME: IPv6 support */
 /* FIXME: The host name lookup may block. We would need an asyncronous
  * get_inaddr function. As a work around, we could let the client do
@@ -737,6 +780,222 @@ int tcp_addr(struct sockaddr_in *sin,
 	return 1;
 }
 
+/* inet_address_info */
+
+static int inet_address2sockaddr_in(struct address_info *c, int sasize, struct sockaddr *sa)
+{
+	CAST(inet_address_info, a, c);
+	struct sockaddr_in *in = (struct sockaddr_in *) sa;
+
+	if (sasize != sizeof(*in))
+		return 0;
+	if (a->ip && !tcp_addr(in,
+				a->ip->length, a->ip->data,
+				a->port))
+		return 0;
+	else
+		if (!a->ip && !tcp_addr(in, 0, NULL, a->port))
+			return 0;
+	return 1;
+}
+
+static int bind_inet_socket(struct address_info *c, int fd)
+{
+	struct sockaddr_in in;
+	int tmp = 1;
+
+	if (inet_address2sockaddr_in(c, sizeof(in), (struct sockaddr *) &in)) {
+		setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &tmp, sizeof(tmp));
+		if (bind(fd, (struct sockaddr *) &in, sizeof(struct sockaddr_in)) == -1) {
+			werror("io.c: bind_inet_socket() bind failed %I:%i %z\n", in.sin_addr, ntohs(in.sin_port), strerror(errno));
+			return 0;
+		}
+	}
+	return 1;
+}
+
+static int connect_inet_socket(struct address_info *c, int fd)
+{
+	struct sockaddr_in in;
+
+	if (inet_address2sockaddr_in(c, sizeof(in), (struct sockaddr *) &in)) {
+		if (connect(fd, (struct sockaddr *) &in, sizeof(in)) < -1) {
+			switch (errno) {
+			case EWOULDBLOCK:
+				return 1;
+			default:
+				return 0;
+			}
+		}
+		return 1;
+	}
+	return 0;
+}
+
+struct inet_address_info *make_inet_address_c(const char *host,
+                                         const char *port)
+{
+	int portno = get_portno(port, "tcp");
+	if (portno < 0)
+		return 0;
+	else
+	{
+		NEW(inet_address_info, info);
+      
+		info->super.family = AF_INET;
+		info->super.bind_socket = bind_inet_socket;
+		info->super.connect_socket = connect_inet_socket;
+		info->port = portno;
+		info->ip = host ? c_format("%z", host) : NULL;
+
+		return info;
+	}
+}
+
+struct inet_address_info *make_inet_address(struct ol_string *host, UINT32 port)
+{
+	NEW(inet_address_info, info);
+
+	info->super.family = AF_INET;
+	info->super.convert2sockaddr = inet_address2sockaddr_in;
+	info->super.bind_socket = bind_inet_socket;
+	info->super.connect_socket = connect_inet_socket;
+	info->port = port; /* htons(port); */
+	info->ip = host;
+	return info;
+}
+
+/* unix_address_info */
+
+static int unix_address2sockaddr_un(struct address_info *c, int sasize, struct sockaddr *sa)
+{
+	CAST(unix_address_info, a, c);
+	struct sockaddr_un *un = (struct sockaddr_un *) sa;
+	if (sasize != sizeof(*un))
+		return 0;
+	strncpy(un->sun_path, a->path->data, MIN(a->path->length + 1, sizeof(un->sun_path)));
+	un->sun_family = AF_UNIX;
+	return 1;
+}
+
+static int bind_unix_socket(struct address_info *c, int fd)
+{
+	struct sockaddr_un un;
+
+	if (unix_address2sockaddr_un(c, sizeof(un), (struct sockaddr *) &un)) {
+		unlink(un.sun_path);
+		errno = 0;
+		if (bind(fd, (struct sockaddr *) &un, sizeof(struct sockaddr_un)) == -1) {
+			werror("io.c: bind_unix_socket(): bind failed %z (%z)\n", 
+			       un.sun_path, strerror(errno));
+			return 0;
+		}
+		return 1;
+	}
+	return 0;
+}
+
+static int connect_unix_socket(struct address_info *c, int fd)
+{
+	struct sockaddr_un un;
+
+	if (unix_address2sockaddr_un(c, sizeof(un), (struct sockaddr *) &un)) {
+		if (connect(fd, (struct sockaddr *) &un, sizeof(un)) < -1) {
+			switch (errno) {
+			case EWOULDBLOCK:
+				return 1;
+			default:
+				return 0;
+			}
+		}
+		return 1;
+	}
+	return 0;
+}
+
+struct unix_address_info *make_unix_address(struct ol_string *path)
+{
+	NEW(unix_address_info, info);
+
+	info->super.family = AF_UNIX;
+	info->super.convert2sockaddr = unix_address2sockaddr_un;
+	info->super.bind_socket = bind_unix_socket;
+	info->super.connect_socket = connect_unix_socket;
+	info->path = path;
+	return info;
+}
+
+struct unix_address_info *make_unix_address_c(const char *path)
+{
+	return make_unix_address(c_format_cstring("%z", path));
+}
+
+struct address_info *sockaddr2address_info(size_t addr_len,
+                                           struct sockaddr *addr)
+{
+	switch(addr->sa_family)
+	{
+	case AF_INET:
+	{
+		struct sockaddr_in *in = (struct sockaddr_in *) addr;
+		UINT32 ip, port;
+
+		ip = ntohl(in->sin_addr.s_addr);
+		port = ntohs(in->sin_port);
+		return &make_inet_address(c_format("%i.%i.%i.%i",
+						  (ip >> 24) & 0xff,
+						  (ip >> 16) & 0xff,
+						  (ip >> 8) & 0xff,
+						  ip & 0xff), port)->super;
+	}
+	case AF_UNIX: {
+		struct sockaddr_un *un = (struct sockaddr_un *) addr;
+		return &make_unix_address_c(un->sun_path)->super;
+	}
+#if 0
+	case AF_INETv6:
+		...
+#endif
+		default:
+			fatal("io.c: sockaddr2info(): Unsupported address family.\n");
+	}
+}
+
+#if 0
+int open_inet_socket(int socktype, int proto, struct sockaddr_in *sa)
+{
+        int fd;
+
+        fd = socket(AF_INET, socktype, proto);
+        if (fd != -1 && sa) {
+        	int tmp = 1;
+                setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &tmp, sizeof(tmp));
+                if (bind(fd, sa, sizeof(struct sockaddr_in)) == -1) {
+			werror("io.c: open_inet_socket() bind failed %z\n", strerror(errno));
+                        close(fd);
+                        fd = -1;
+                }
+        }
+        return fd;
+}
+#endif
+
+int open_socket(int family, int socktype, int proto, struct address_info *a)
+{
+
+	int fd;
+	fd = socket(family, socktype, proto);
+	if (fd != -1 && a) {
+		if (a && !ADDRESS_BIND(a, fd)) {
+			werror("open_socket(): error binding socket\n");
+			close(fd);
+			return -1;
+		}
+
+	}
+	return fd;
+}
+
 /* For fd:s in blocking mode. */
 int write_raw(int fd, UINT32 length, UINT8 *data)
 {
@@ -823,20 +1082,33 @@ void io_init_fd(int fd)
   io_set_close_on_exec(fd);
 }
 
+int io_open_socket(int family, int socktype, int proto, struct address_info *local)
+{
+	int s;
+
+	s = socket(family, socktype, proto);
+	if (s < 0)
+		return -1;
+  	io_init_fd(s);
+	if (local && ADDRESS_BIND(local, s))
+		return s;
+	else
+		if (!local)
+			return s;
+	close(s);
+	return -1;
+}
+
 /* Some code is taken from bellman's tcputils. */
 struct connect_fd *io_connect(struct io_backend *b,
 			      int s,
-			      struct sockaddr_in *remote,
+			      struct address_info *remote,
 			      struct fd_callback *f)
 {
 #if 0
 	debug("io.c: connecting using fd %i\n", s);
 #endif
-  
-	io_init_fd(s);
-
-	if ( (connect(s, (struct sockaddr *)remote, sizeof *remote) < 0)
-	     && (errno != EINPROGRESS) ) {
+	if (!ADDRESS_CONNECT(remote, s)) {
 		int saved_errno = errno;
 		close(s);
 		errno = saved_errno;
@@ -860,11 +1132,10 @@ struct listen_fd *io_listen(struct io_backend *b,
 			    int s,
 			    struct fd_listen_callback *callback)
 {
+	if (s < 0) return NULL;
 #if 0
 	debug("io.c: listening on fd %i\n", s);
 #endif
-  
-	io_init_fd(s);
 
 	if (listen(s, 256) < 0) {
 		close(s);
